/*******************************************************************
* SNAAB ILI9341 Arduino Library for SPI ILI9341 TFT Color Display
* version 1.0.0
* distributed under GPL v2.0
* Copyright (c) 2007-2021, SNAAB Systems 
* www.snaabsys.ru
*
* uses part of Stanley Huang (http://arduinodev.com) library code
*******************************************************************/

#ifndef SNAAB_ILI9341_H
#define SNAAB_ILI9341_H

#include "Arduino.h"
#include <SPI.h>

/* 
Pins HW SPI Arduino Nano
D13 : SCK
D12 : MISO
D11 : MOSI

Default pins for ILI9341
D10 : CS
D9  : D/C
*/

#define LANDSCAPE       0
#define PORTRAIT        1
#define LANDSCAPEROTATE 2
#define PORTRAITROTATE  3

// Color definitions
#define BLACK       0x0000    ///<   0,   0,   0 
#define WHITE       0xFFFF    ///< 255, 255, 255
#define RED         0xF800    ///< 255,   0,   0 
#define GREEN       0x07E0    ///<   0, 255,   0 
#define BLUE        0x001F    ///<   0,   0, 255 
#define YELLOW      0xFFE0    ///< 255, 255,   0 
#define MAGENTA     0xF81F    ///< 255,   0, 255 
#define CYAN        0x07FF    ///<   0, 255, 255 
#define ORANGE      0xFD20    ///< 255, 165,   0 
#define GREENYELLOW 0xAFE5    ///< 173, 255,  47 
#define LIGHTGREY   0xC618    ///< 198, 198, 198 
#define DARKGREY    0x7BEF    ///< 128, 128, 128 
#define MAROON      0x7800    ///< 128,   0,   0 
#define DARKGREEN   0x03E0    ///<   0, 128,   0 
#define NAVY        0x000F    ///<   0,   0, 128 
#define OLIVE       0x7BE0    ///< 128, 128,   0 
#define PURPLE      0x780F    ///< 128,   0, 128 
#define DARKCYAN    0x03EF    ///<   0, 128, 128 
#define PINK        0xFC18    ///< 255, 130, 198

//TFT resolution 240*320
#define MIN_X	0
#define MIN_Y	0
#define MAX_X	239
#define MAX_Y	319

#define TFT_CS_LOW  digitalWrite(PIN_CS, LOW)
#define TFT_CS_HIGH digitalWrite(PIN_CS, HIGH)
#define TFT_DC_LOW  digitalWrite(PIN_DC, LOW)
#define TFT_DC_HIGH digitalWrite(PIN_DC, HIGH)  

#define lcMax  32
#define lcStep  8


/*************************************************************************
* The ListChar structure contains a compressed list of available
* characters in the font. 
* Reduce the font size by removing unused characters from the font in the
* SNAAB FontEditor (www.snaabsys.ru)
*************************************************************************/
typedef struct {
  uint8_t listofchar[lcMax]; // list of available symbols
} ListChar;

/*************************************************************************
* The CharInfo structure contains information about the font character.
* For more details, see SNAAB FontEditor (www.snaabsys.ru)
*************************************************************************/
typedef struct {
  uint8_t  code;          // current code 
  uint8_t  topoffset;     // top offset of character in pixels
  uint8_t  realwidth;     // real width of character in pixels
  uint8_t  realheight;    // real height of character in pixels
  uint8_t  space;         // distance between characters in pixels  
  uint16_t blocksize;     // size in bytes of character
  uint16_t blockstart;    // start of characters
} CharInfo;

/*************************************************************************
* The CurrentFont structure contains information about the font. 
*************************************************************************/
typedef struct {
  uint8_t  charcount;     // characters count
  uint8_t  width;         // width of font in pixels
  uint8_t  height;        // height of font in pixels
  uint8_t  ascent;        // ascent of font in pixels
  uint8_t  descen;        // descent of font in pixels			
  ListChar *avchar;       // available characters
  uint8_t  *data;         // data of char
  CharInfo *info;         // information on characters  	
} CurrentFont;

class SNAAB_ILI9341 {
  public:
	void begin (uint8_t CS, uint8_t DC);
	void clear(void);
    void setFreq(int devfreq);	
	void setOrientation(uint8_t val);

	void fillScreen(uint16_t color);
	
	void drawPixel(uint16_t X0, uint16_t Y0, uint16_t color);
		
	void drawHorLine(uint16_t XL, uint16_t YL, uint16_t length, uint16_t color);		
	void drawVertLine(uint16_t XL, uint16_t YL, uint16_t length, uint16_t color);			
	void drawLine(uint16_t X0, uint16_t Y0, uint16_t X1, uint16_t Y1, uint16_t color);
	
	void drawRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t color);
	void drawThickRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t thick, uint16_t color);
	
	void drawRoundRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t r,  uint16_t color);		
	void drawThickRoundRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t thick, uint16_t r,  uint16_t color);	
	
	void fillRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t color);
	void fillRoundRect(uint16_t XL, uint16_t YL, uint16_t width, uint16_t height, uint16_t r, uint16_t color);
	void drawCircle(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t color);
    void fillCircle(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t color);
			
	void drawRing(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t thick, uint16_t color);		
			
	void setFont(const CurrentFont *font = NULL);		
	
	uint16_t drawChar(uint16_t ct, int16_t X0, int16_t Y0, bool baseline, uint16_t color);
	uint16_t drawChar(const CurrentFont *font, uint16_t ct, int16_t X0, int16_t Y0, bool baseline, uint16_t color);

	uint16_t drawText(char *str, int16_t X0, int16_t Y0, bool baseline, bool mode, uint16_t color);	
	uint16_t drawText(const CurrentFont *font, char *str, int16_t X0, int16_t Y0, bool baseline, bool mode, uint16_t color);
		
	void fillChar(char *str, int16_t X0, int16_t Y0, bool baseline, uint16_t index, uint16_t color);
	void fillChar(const CurrentFont *font, char *str, int16_t X0, int16_t Y0, bool baseline, uint16_t index, uint16_t color);
		
	void drawBitmap(const uint8_t bitmap[], int16_t XL, int16_t YL, int16_t w, int16_t h, uint16_t color);
		
  private:
/*
Default pins for ILI9341
D10 : CS
D9  : D/C
*/  
    uint8_t PIN_CS = 10;
    uint8_t PIN_DC = 9;
 
	void sendCMD(uint8_t index);
	void setCol(uint16_t StartCol,uint16_t EndCol);
	void setPage(uint16_t StartPage,uint16_t EndPage);
	void WRITE_DATA(uint8_t data);	
	void sendData(uint16_t data);	
	uint8_t Read_Register(uint8_t Addr,uint8_t xParameter);
	uint8_t readID(void);
    void setXY(uint16_t x0, uint16_t x1, uint16_t y0, uint16_t y1);	
	void doFillRect(uint16_t XL, uint16_t XR,uint16_t YT, uint16_t YB, uint16_t color);	
	
	void doDrawCorner(uint16_t X0, uint16_t Y0, uint16_t r, uint8_t a, uint16_t color);
	void doDrawThickCorner(uint16_t X0, uint16_t Y0, uint16_t r, uint16_t delta, uint8_t a, uint16_t color);
    void doFillCorner(uint16_t X0, uint16_t Y0, uint16_t r, uint8_t a, uint16_t color);	

	String UTF8to1251(char *str);	
	unsigned char UTF8to1251(uint16_t ct);	

	uint8_t Orientation = 0;
	
	ListChar *lChar = NULL;
	CharInfo *cInfo = NULL;
	CurrentFont *cFont = NULL;
	uint8_t  *fData = 0;
		
	uint8_t  fcharcount = 0; // characters count
    uint8_t  fwidth = 0;     // width of font in pixels    
    uint8_t  fheight = 0;    // height of font in pixels
	uint8_t  fascent = 0;    // ascent of font in pixels
};

#endif //SNAAB_ILI9341_H
